import sys
import polars as pl
from typing import Union
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController

def sort(df: pl.LazyFrame, colnames: Union[str, tuple[str], list[str]], desc: bool = False) -> pl.LazyFrame:
    """[chainable] Sorts all rows by the specified column values."""
    # prevent type guessing
    colnames: tuple[str]
    if type(colnames) is list:
        colnames = tuple(colnames)
    elif type(colnames) is str:
        colnames = (colnames, )

    if not exists_colname(df=df, colnames=colnames):
        sys.exit(1)

    LogController.debug(f"sort by {colnames} ({'desc' if desc else 'asc'}).")
    return df.sort(colnames, descending=desc)



import sys
import polars as pl
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController

def isin(df: pl.LazyFrame, colname: str, values: list) -> pl.LazyFrame:
    """[chainable] Filters rows containing the specified values."""
    LogController.debug(f"filter condition: {values} in {colname}")
    if not exists_colname(df=df, colnames=[colname]):
        sys.exit(1)

    return df.filter(pl.col(colname).is_in(values))



import io
from qsv import main
from qsv.controllers.DataFrameController import DataFrameController
import polars

scriptpath = '__init__.py'
logpath = 'Security.csv'

def test_load_unit():
    d = DataFrameController()
    d.load(logpath)
    assert type(d.df) == polars.lazyframe.frame.LazyFrame

def test_load_cli(monkeypatch):
    argv = [scriptpath, "load", logpath]
    stdin = io.StringIO()
    stdout = io.StringIO()

    with monkeypatch.context() as m:
        m.setattr("sys.argv", argv)
        m.setattr("sys.stdin", stdin)
        m.setattr("sys.stdout", stdout)
        main()
        assert stdout.getvalue().startswith('shape: ')

def test_load_quilt(tmpdir, monkeypatch):
    config = """title: test
rules:
    load:
"""
    temp_file = tmpdir.join('tmprule.yaml')
    temp_file.write(config)

    argv = [scriptpath, "quilt", str(temp_file), logpath]
    stdin = io.StringIO()
    stdout = io.StringIO()

    with monkeypatch.context() as m:
        m.setattr("sys.argv", argv)
        m.setattr("sys.stdin", stdin)
        m.setattr("sys.stdout", stdout)
        main()
        assert stdout.getvalue().startswith('Loaded Rules')


import sys
from pathlib import Path

from qsv.controllers.LogController import LogController

def exists_path(path: tuple[str]) -> bool:
    for p in path:
        if not Path(p).exists():
            LogController.error(f"[Error] File \"{p}\" does not exist.")
            return False
    else:
        return True



from pathlib import Path
import polars as pl

class CsvController(object):
    def __init__(self, path):
        self.path: Path = path

    def get_dataframe(self, separator: str = ',', low_memory: bool = False) -> pl.DataFrame:
        df = pl.scan_csv(
            self.path,
            try_parse_dates=True,
            rechunk=True,
            truncate_ragged_lines=True,
            separator=separator,
            low_memory=low_memory,
        )
        return df



import sys
import polars as pl
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController

def changetz(df: pl.LazyFrame, colname: str, tz_from: str = "UTC", tz_to: str = "UTC", dt_format: str = None) -> pl.LazyFrame:
    """[chainable] Changes the timezone of the specified date column."""
    # convert string to datetime
    if df.select(colname).collect_schema().dtypes()[0] != pl.Datetime:
        if dt_format:
            df = df.with_columns(pl.col(colname).str.to_datetime(dt_format))
        else:
            df = df.with_columns(pl.col(colname).str.to_datetime())

    if not exists_colname(df=df, colnames=[colname]):
        sys.exit(1)

    LogController.debug(f"change {colname} timezone {tz_from} to {tz_to}.")

    # setup and change timezone
    df = df.with_columns(pl.col(colname).dt.replace_time_zone(tz_from))
    df = df.with_columns(pl.col(colname).dt.convert_time_zone(tz_to))
    return df



import polars as pl
from qsv.controllers.LogController import LogController

def head(df: pl.LazyFrame, number: int = 5) -> pl.LazyFrame:
    """[chainable] Selects only the first N lines."""
    LogController.debug(f"heading {number} lines.")
    return df.head(number)



from qsv.operations.chainables.select import select
from qsv.operations.chainables.isin import isin
from qsv.operations.chainables.contains import contains
from qsv.operations.chainables.sed import sed
from qsv.operations.chainables.grep import grep
from qsv.operations.chainables.head import head
from qsv.operations.chainables.tail import tail
from qsv.operations.chainables.sort import sort
from qsv.operations.chainables.uniq import uniq
from qsv.operations.chainables.changetz import changetz
from qsv.operations.chainables.renamecol import renamecol

__all__ = [
    'select',
    'isin',
    'contains',
    'sed',
    'grep',
    'head',
    'tail',
    'sort',
    'uniq',
    'changetz',
    'renamecol',
]


from rich import box
from rich.table import Table
from rich.console import Console


class TableView(object):
    @staticmethod
    def print(headers: list[str], values: list[list[str]]):
        table = Table(show_header=True, box=box.SQUARE_DOUBLE_HEAD)

        for header in headers:
            table.add_column(header)
        
        for value in values:
            table.add_row(*value)

        console = Console()
        console.print(table)


import sys
import polars as pl
from typing import Union
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController

def uniq(df: pl.LazyFrame, colnames: Union[str, tuple[str], list[str]]) -> pl.LazyFrame:
    """[chainable] Remove duplicate rows based on the specified column names."""
    # prevent type guessing
    colnames: tuple[str]
    if type(colnames) is list:
        colnames = tuple(colnames)
    elif type(colnames) is str:
        colnames = (colnames, )

    if not exists_colname(df=df, colnames=colnames):
        sys.exit(1)

    LogController.debug(f"unique by {colnames}.")
    return df.unique(subset=colnames)



import polars as pl

def grep(df: pl.LazyFrame, pattern: str, ignorecase: bool = False) -> pl.LazyFrame:
    """[chainable] Treats all columns as strings and filters rows where any column matches the specified regex."""
    df = df.with_columns(
        pl.concat_str(
            [pl.col(colname).cast(pl.String).fill_null("") for colname in df.collect_schema().names()],
            separator=","
        ).alias('___combined')
    )
    df = df.filter(
        pl.col('___combined').str.contains(f"(?i){pattern}") if ignorecase else pl.col('___combined').str.contains(pattern)
    )
    df = df.drop(['___combined'])
    return df



import sys
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController
import polars as pl

def contains(df: pl.LazyFrame, colname: str, pattern: str, ignorecase: bool = False) -> pl.LazyFrame:
    """[chainable] Filters rows where the specified column matches the given regex."""
    if not exists_colname(df=df, colnames=[colname]):
        sys.exit(1)
    
    LogController.debug(f"filter condition: {pattern} contains {colname}")
    pattern = pattern if type(pattern) is str else str(pattern)
    return df.filter(
        pl.col(colname).str.contains(f"(?i){pattern}") if ignorecase else pl.col(colname).str.contains(pattern)
    )



import sys
from qsv.utils.DataFrameUtils import exists_colname
import polars as pl

def renamecol(df: pl.LazyFrame, colname: str, new_colname: str) -> pl.LazyFrame:
    """[chainable] Renames the specified column."""
    if not exists_colname(df=df, colnames=[colname]):
        sys.exit(1)

    return df.rename({colname: new_colname})



import sys
import polars as pl
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController

def sed(df: pl.LazyFrame, colname: str, pattern: str, replacement: str, ignorecase: bool = False) -> pl.LazyFrame:
    """[chainable] Replaces values using the specified regex."""
    if not exists_colname(df=df, colnames=[colname]):
        sys.exit(1)

    LogController.debug(f"sed condition: {pattern} on {colname}")
    pattern = pattern if type(pattern) is str else str(pattern)
    return df.with_columns(
        pl.col(colname).cast(pl.String).str.replace(f"(?i){pattern}", replacement) if ignorecase else pl.col(colname).cast(pl.String).str.replace(pattern, replacement)
    )



import polars as pl
from qsv.controllers.LogController import LogController

def exists_colname(df: pl.LazyFrame, colnames: list[str]) -> bool:
    columns = df.collect_schema().names()
    for colname in colnames:
        if colname not in columns:
            LogController.error(f"Column \"{colname}\" does not exist in the dataframe.")
            return False
    else:
        return True



import sys
from typing import Union
from qsv.utils.DataFrameUtils import exists_colname
from qsv.controllers.LogController import LogController
import polars as pl

def parse_columns(headers: list[str], colnames: tuple[str]) -> list[str]:
    parsed_columns = list()
    for col in colnames:
        if '-' in col:
            # parse 'startcol-endcol' string
            flag_extract = False
            start, end = col.split('-')
            for h in headers:
                if h == start:
                    flag_extract = True
                if flag_extract:
                    parsed_columns.append(h)
                if h == end:
                    flag_extract = False
        else:
            parsed_columns.append(col)
    return parsed_columns


def select(df: pl.LazyFrame, colnames: Union[str, tuple[str]]) -> pl.LazyFrame:
    """[chainable] Selects only the specified columns."""

    # prevent type guessing
    colnames: tuple[str]
    if type(colnames) is list:
        colnames = tuple(colnames)
    elif type(colnames) is str:
        colnames = (colnames, )

    if not exists_colname(df=df, colnames=colnames):
        sys.exit(1)

    selected_columns = parse_columns(headers=df.collect_schema().names(), colnames=colnames)
    LogController.debug(msg=f"{len(selected_columns)} columns are selected. {', '.join(selected_columns)}")
    return df.select(selected_columns)



import polars as pl
from qsv.controllers.LogController import LogController

def tail(df: pl.LazyFrame, number: int = 5) -> pl.LazyFrame:
    """[chainable] Selects only the last N lines."""
    LogController.debug(f"tailing {number} lines.")
    return df.tail(number)



import sys
from qsv.utils.FileUtils import exists_path
from qsv.controllers.CsvController import CsvController
from qsv.controllers.LogController import LogController

def load(path: tuple[str], separator: str = ',', low_memory: bool = False):
    """[initializer] Loads the specified CSV files."""
    if not exists_path(path=path):
        sys.exit(1)

    LogController.debug(msg=f"{len(path)} files are loaded. [{', '.join(path)}]")
    return CsvController(path=path).get_dataframe(separator=separator, low_memory=low_memory)



from qsv.operations.finalizers.headers import headers
from qsv.operations.finalizers.stats import stats
from qsv.operations.finalizers.showquery import showquery
from qsv.operations.finalizers.show import show
from qsv.operations.finalizers.showtable import showtable
from qsv.operations.finalizers.dump import dump

__all__ = [
    'headers',
    'stats',
    'showquery',
    'show',
    'showtable',
    'dump',
]


import yaml
from pathlib import Path

# for allow duplicate keynames.
def rename_duplicate_keys(loader, node, deep=False):
    mapping = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        value = loader.construct_object(value_node, deep=deep)

        new_key = key
        while new_key in mapping:
            new_key += "_"

        mapping[new_key] = value

    return mapping

class YamlController(object):
    def __init__(self):
        yaml.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            rename_duplicate_keys,
            Loader=yaml.SafeLoader
        )

    def load_data(self, path):
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
            return data
        


from qsv.views.TableView import TableView
import polars as pl

def headers(df=pl.LazyFrame, plain: bool = False) -> None:
    """[finalizer] Displays the column names of the data."""
    if plain:
        print(",".join([f"\"{c}\"" for c in df.collect_schema().names()]))
    else:
        digits = len(str(len(df.collect_schema().names())))
        TableView.print(
            headers=[f"{''.join([' ' for _ in range(0, digits-1)])}#", "Column Name"],
            values=[[str(i).zfill(digits), c] for i, c in enumerate(df.collect_schema().names())]
        )



import logging

class LogController(object):
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s [%(levelname)s] %(message)s',
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        handlers=[
            logging.StreamHandler()
        ]
    )
    logger = logging.getLogger(__name__)
    logger.disabled = True

    def info(msg: str) -> None:
        LogController.logger.info(msg)

    def debug(msg: str) -> None:
        LogController.logger.debug(msg)

    def warning(msg: str) -> None:
        LogController.logger.warning(msg)

    def error(msg: str) -> None:
        LogController.logger.error(msg)



from pathlib import Path
from qsv.views.TableView import TableView
from qsv.controllers.YamlController import YamlController

class QuiltController(object):
    def __init__(self):
        pass

    def load_configs(self, config) -> list[dict]:
        configs = list()
        if Path(config).is_dir():
            # I think the yaml format is great, the only drawback is that it has two extensions, .yaml and .yml.
            for f in Path(config).glob("**/*.yaml"):
                configs.append(YamlController().load_data(f))
            for ff in Path(config).glob("**/*.yml"):
                configs.append(YamlController().load_data(ff))
        else:
            configs.append(YamlController().load_data(Path(config)))

        return configs

    def print_configs(self, configs: list[dict]):
        print(f"Loaded {len(configs)} Rules")
        digits = len(str(len(configs)))
        TableView.print(
            headers=[
                f"{''.join([' ' for _ in range(0, digits-1)])}#",
                'title',
                'description',
                'version',
                'author'
            ],
            values=[
                [
                    str(i).zfill(digits),
                    c.get('title'),
                    c.get('description'),
                    c.get('version'),
                    c.get('author')
                ] for i, c in enumerate(configs, 1)
            ]
        )



import re
from typing import Union, Optional, Literal
from qsv.controllers.LogController import LogController
from qsv.controllers.QuiltController import QuiltController
import polars as pl

def process(dfc, steps: dict, path: tuple[str], source: Optional[pl.LazyFrame] = None) -> pl.LazyFrame:
    default_path = path

    # drop dataframe before each stage
    dfc.drop()

    # pipeline
    if source is not None:
        dfc.df = source

    for k, v in steps:
        # for allow duplicated rulenames.
        k = k if not k.endswith('_') else re.sub(r'_+$', '', k) 

        if k == 'load':
            dfc.load(*path)
        elif v:
            getattr(dfc, k)(**v)
        else:
            getattr(dfc, k)()
    
    return dfc.df


def concat(
        df_list: list[pl.LazyFrame],
        how: Literal['vertical', 'vertical_relaxed', 'horizontal', 'diagonal', 'align']
    ) -> pl.LazyFrame:
    return pl.concat(df_list, how=how)


def join(
        df_list: list[pl.LazyFrame],
        how: Literal['inner', 'left', 'right', 'full', 'semi', 'anti', 'cross'],
        key: Union[str, list[str]],
        coalesce: bool
    ) -> pl.LazyFrame:
    return df_list[0].join(df_list[1], how=how, on=key, coalesce=coalesce)


def quilt(dfc, config: str, path: tuple[str]) -> None:
    """[quilter] Loads the specified quilt batch files."""
    LogController.debug(f"config: {config}")
    LogController.debug(f"{len(path)} files are loaded. [{', '.join(path)}]")

    q = QuiltController()
    configs = q.load_configs(config)
    q.print_configs(configs)

    # per config file
    for config in configs:
        # per stage
        df_dict = dict()

        for stage_key, stage_values in config.get('stages').items():
            stage_type = stage_values.get('type')

            if stage_type == "process":
                if 'source' in stage_values:
                    # pipeline
                    df_dict[stage_key] = process(
                        dfc=dfc,
                        steps=stage_values.get('steps').items(),
                        path=path,
                        source=df_dict[stage_values.get('source')],
                    )
                else:
                    # load from file
                    df_dict[stage_key] = process(
                        dfc=dfc,
                        steps=stage_values.get('steps').items(),
                        path=path
                    )

            elif stage_type == "concat":
                sources = stage_values.get('sources')
                how = stage_values.get('params', dict()).get('how', 'vertical')
                df_dict[stage_key] = concat(
                    df_list=[df_dict.get(source) for source in sources],
                    how=how
                )

            elif stage_type == "join":
                sources = stage_values.get('sources')
                how = stage_values.get('params', dict()).get('how', 'full')
                key = stage_values.get('params', dict()).get('key')
                coalesce = stage_values.get('params', dict()).get('coalesce', True)

                df_dict[stage_key] = join(
                    df_list=[df_dict.get(source) for source in sources],
                    how=how,
                    key=key,
                    coalesce=coalesce
                )



from typing import Union

from qsv.operations.initializers import *
from qsv.operations.chainables import *
from qsv.operations.finalizers import *
from qsv.operations.quilters import *


class DataFrameController(object):
    def __init__(self):
        self.df = None
        
    # -- initializers --
    def load(self, *path: tuple[str], separator: str = ',', low_memory: bool = False):
        self.df = load(path=path, separator=separator, low_memory=low_memory)
        return self

    # -- chainables --
    def select(self, colnames: Union[str, tuple[str]]):
        self.df = select(df=self.df, colnames=colnames)
        return self
    
    def isin(self, colname: str, values: list):
        self.df = isin(df=self.df, colname=colname, values=values)
        return self
    
    def contains(self, colname: str, pattern: str, ignorecase: bool = False):
        self.df = contains(df=self.df, colname=colname, pattern=pattern, ignorecase=ignorecase)
        return self

    def sed(self, colname: str, pattern: str, replacement: str, ignorecase: bool = False):
        self.df = sed(df=self.df, colname=colname, pattern=pattern, replacement=replacement, ignorecase=ignorecase)
        return self
    
    def grep(self, pattern: str, ignorecase: bool = False):
        self.df = grep(df=self.df, pattern=pattern, ignorecase=ignorecase)
        return self

    def head(self, number: int = 5):
        self.df = head(df=self.df, number=number)
        return self

    def tail(self, number: int = 5):
        self.df = tail(df=self.df, number=number)
        return self
    
    def sort(self, colnames: Union[str, tuple[str], list[str]], desc: bool = False):
        self.df = sort(df=self.df, colnames=colnames, desc=desc)
        return self
    
    def uniq(self, colnames: Union[str, tuple[str], list[str]]):
        self.df = uniq(df=self.df, colnames=colnames)
        return self
    
    def changetz(self, colname: str, tz_from: str = "UTC", tz_to: str = "UTC", dt_format: str = None):
        self.df = changetz(df=self.df, colname=colname, tz_from=tz_from, tz_to=tz_to, dt_format=dt_format)
        return self

    def renamecol(self, colname: str, new_colname: str):
        self.df = renamecol(df=self.df, colname=colname, new_colname=new_colname)
        return self

    def drop(self):
        self.df = None
        return self
    
    # -- finalizer --
    def headers(self, plain: bool = False) -> None:
        headers(df=self.df, plain=plain)
    
    def stats(self) -> None:
        stats(df=self.df)

    def showquery(self) -> None:
        showquery(df=self.df)

    def show(self) -> None:
        show(df=self.df)

    def showtable(self) -> None:
        showtable(df=self.df)
    
    def dump(self, path: str = None) -> None:
        dump(df=self.df, path=path)

    # -- quilter --
    def quilt(self, config: str, *path: tuple[str]) -> None:
        quilt(dfc=self, config=config, path=path)
    
    def __str__(self):
        if self.df is not None:
            print(self.df.collect())
        return ''



import re
from datetime import datetime
from qsv.controllers.LogController import LogController
import polars as pl

def autoname(df: pl.LazyFrame) -> str:
    now = datetime.now().strftime('%Y%m%d-%H%M%S')
    query = df.explain(optimized=False).splitlines()[0]
    temp = re.sub(r'[^\w\s]', '-', query)
    temp = re.sub(r'-+', '-', temp)
    temp = temp.strip('-')
    temp = temp.replace(' ', '')
    temp = temp.lower()
    return f"{now}_{temp}.csv"


def dump(df: pl.LazyFrame, path: str = None) -> None:
    """[finalizer] Outputs the processing results to a CSV file."""
    path = path if path else autoname(df=df)
    df.collect().write_csv(path)
    LogController.info(f"csv dump successfully: {path}")
